/*
 * rfft.c
 */

#include "twcw.h"
#include "common.h"

/* Forward decl's */
float average (int cnt, float times[], int *curr);

/* Globals */
rfftw_plan plan;               /* plan for FFTW */
int sens = DA_HEIGHT/2;        /* sensitivity */
float wordsf = 0.0;            /* make this a resource someday */
float charsf = 0.0;            /* make this a resource someday */


/*
 * init_fftw - create a plan for FFTW
 */
void init_plan (void)
{
   plan = rfftw_create_plan (SAMPLE_SIZE, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);
}


/*
 * destroy_plan - create a plan for FFTW
 */
void destroy_plan (void)
{
   rfftw_destroy_plan (plan);
}


/*
 * set_sens
 */
void set_sens (int val)
{
   sens = DA_HEIGHT * val / 100;
   // printf ("sens %d\n", sens);
}


/*
 * set_wordsf
 */
void set_wordsf (int val)
{
   //wordsf = val / 10.0;
   wordsf = val * 0.05;
   // printf ("wordsf %f\n", wordsf);
}


/*
 * set_charsf
 */
void set_charsf (int val)
{
   //charsf = (val + 5) * 0.1;
   charsf = val / 20.0;
   // printf ("charsf %f\n", charsf);
}


/*
 * doRFFTW - wrapper for FFTW. Converts sample
 * to fftw_real type, runs the FFT, generates
 * the power array and scales it
 */
void doRFFTW (unsigned int sample[], fftw_real power[])
{
   int i;
   float sf = 2e-7;
   fftw_real in[SAMPLE_SIZE], out[SAMPLE_SIZE];

   for (i=0; i<SAMPLE_SIZE; i++)
   {
      in[i] = (fftw_real) ((char)( sample[i]));
   }

   rfftw_one (plan, in, out);

   // TJW power[0] = (out[0] * out[0]) * sf;  /*DC part */
   power[0] = 0;

   for (i=5; i<(SAMPLE_SIZE)/2; i++)
   {
      power[i] = (out[i] * out[i] +
                   out[SAMPLE_SIZE-i] * out[SAMPLE_SIZE-i]) * sf;
   }

   if (SAMPLE_SIZE % 2 == 0)  /* sample == even */
   {
      power[SAMPLE_SIZE/2] = (out[SAMPLE_SIZE/2] * out[SAMPLE_SIZE/2]) * sf;
   }
}


/*
 * decode
 */
#define MEM_SIZE 32

void decode (fftw_real power[], int lo, int hi)
{
   int i, sigs = 0;
   // TJW fftw_real level = 0.0;

   char str[5];   //temp

   char letter[3];
   static int did_space, did_letter;
   static int markcnt, markindex;
   static int spacecnt; // TJW, spaceindex;
   static float markavg, markmem[MEM_SIZE];
   // TJW static float spaceavg, spacemem[MEM_SIZE];
   /*
    * First, pre-screen. See how many have power in fft RX_FREQ bins
    * greater than sens level
    */
   for (i=lo; i<=hi; i++)        
   {
      // TJW level = level + power[i];
      if (power[i] > sens)
         sigs++;
   }

   /*
    * If this is true, we are in a mark, so count it and return
    */
   // TJW if (level / (hi-lo) > sens )       /* good ones are > sens */
   if (sigs >= 2)                     /* got good ones > sens */
   {
      markcnt++;                      /* got a mark so count it and */
      did_space = False;              /* setup the control flags */
      did_letter = False;
      return;                         /* go back for the next sample */
   }

   /*
    * If this is true, we found the first space right after a mark, so...
    * calculate the space and average mark averages, build the morse
    * character string and count this space.
    */
   if (markcnt > 0)    /* or markcnt > 0.5 average for noise? got the time? */
   {
      //spaceavg = average ( spacecnt, spacemem, &spaceindex);
      markavg = average (markcnt, markmem, &markindex);

#if 1
sprintf (str, "%1.1f", markavg/2.0);
XtVaSetValues (tf1,
   XmNvalue, str,
   NULL);
sprintf (str, "%1.1f", markavg);
XtVaSetValues (tf2,
   XmNvalue, str,
   NULL);
#endif

      /* convert the last mark to morse '.' or '-' string */
      if (markcnt < markavg)
      {
         next_symbol ('.');
      }
      else
      {
         next_symbol ('-');
      }

      spacecnt = 1;                     /* count this space */
      markcnt = 0;                      /* get ready for the next mark */
      return;                           /* go back for the next sample */
   }

   /*
    * If we get here, we are in a space so count the space.
    * Skip it if we have already printed the space character, so
    * dead air won't skew the space average. 
    */
   if (did_space == False)
   {
      spacecnt++;

      // if ((spacecnt >= (float)spaceavg * charsf) && (did_letter == False))
      if (spacecnt >= markavg - (charsf * markavg) && did_letter == False)
      {
         letter[0] = get_letter();
         letter[1] = '\0';
         XmTextInsert (rxText, XmTextGetLastPosition(rxText), letter);
         did_letter = True;
         return;
      }
      else
      {
         //spacecnt++;
      }

      if (spacecnt >= 3.0 * (markavg + (wordsf * markavg) ))
      {
#if 1
sprintf (str, "%d", spacecnt);
XtVaSetValues (tf3,
   XmNvalue, str,
   NULL);
#endif
         XmTextInsert (rxText, XmTextGetLastPosition(rxText), "  " );
         did_space = True;
      }
   }
}


/*
 * average for marks
 * Computes and returns the average of an array of MEM_SIZE.
 * // if last 3 are the same (/+\- some) don't include the current count
 * Also increments the index for the array and recirulates
 * at MEM_SIZE
 */
float average (int cnt, float times[], int *index)
{
   int i, k=0;
   float avg = 0;
   int last = 0; 
   static int skew;

   if (*index == 0)
   {
      last = 31;
   }
   else
   {
      last = *index - 1;
   }

/* if this cnt and the previous 2 are ~=, don't include it in the average. */
   if (cnt > (times[last] - 0.12 * times[last]) &&
       cnt < (times[last] + 0.12 * times[last]))
   {
      skew++;
   }

   if ( skew < 3 )
   {
      times[*index] = (float)cnt;
   }
   else
   {
      skew = 0;
   }
      /* do the average */
      for (i=0; i<MEM_SIZE; i++)
      {
         if (times[i] == 0.0)  /* speeds up getting average on startup */
            break;
         avg = avg + times[i];
         k++;
      }
   avg = avg / k;

   if (skew < 3 )
   {
      (*index)++;
      if (*index == MEM_SIZE)  /* loop the index */ 
      {
         *index = 0;
      }
   }
   return (avg);
}

