/*
 * twcw:  A gui application that sends Morse Code 
 * Copyright (C) 1997 Ted Williams WA0EIR 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 * Versions: 1.2 -  APR 2004
 */

#include "common.h"
#include "sendCW.h"

int arg = TIOCM_RTS;
int beepPC = 0;
int beepSC = 0;


/*
 * DoWord FUNCTION
 */
void doWord (unsigned char w[])
{
   int  i = 0;
   struct tag1 reply;

   while (w[i] != '\0')
   {
      if (letters[(int) w[i]] == (char) NULL)  /* Unknown letter, so skip it */
      {
         i++;
         continue;
      }
      /* Should we send this letter via cwlib? */
      if (beepSC == 1 || beepPC == 1)
      {
         cw_send_character (w[i]);
      }

      doLetter (letters[w[i]]);

      /* Send the letter back to the GUI? */
      reply.type = SENT_MSG;
      reply.data.word[0] = w[i];
      reply.data.word[1] = (char) NULL;
      if (msgsnd (qid, (struct msgbuf *)&reply, sizeof(reply.data), 0) == -1)
      {
         perror ("keyer.c doWord: msgsnd SENT_MSG failed");
         exit (1);
      }
      i++;
   }

   timer (iw_time);         /* inter word time */

   /* Done with a word, so send a space to GUI */
   reply.type = SENT_MSG; 
   reply.data.word[0] = w[i-1];
   reply.data.word[1] = (char) NULL; 
 
   if (msgsnd (qid, (struct msgbuf *) &reply, sizeof (reply.data), 0) == -1)
   {
      perror ("keyer.c doWord: msgsnd SENT_MSG space ");
      exit (1);
   }
}


/*
 * doLetter FUNCTION
 */
void doLetter (char ch[])
{
   int i = 0;

   while (ch[i] != '\0')
   {
      if (ch[i] == DIT)
         keyRTS (dit_time);
      if (ch[i] == DAH)
         keyRTS (dah_time);
      i++;
   }
   timer (il_time);          /* inter letter time */
}


/*
 * keyRTS FUNCTION
 * keys RTS and cwirc in_key
 */
void keyRTS (int time)
{
   if (key_mode == CWIRC)
   {
      cwirc_api_shm->in_key = 1;
   }
   else  /* must be TWCW mode so key the serial card on */
   {
      if (ioctl (serfd, TIOCMBIS, &arg) == -1)
      {
         perror("keyer.c keyRTS: RTS Set failed ");
      }
   }

   timer (time);                         /* dit or dah time */

   if (key_mode == CWIRC)
   {
      cwirc_api_shm->in_key = 0;
   }
   else   /* must be TWCW mode so key serial card off */
   {
      if (ioctl (serfd, TIOCMBIC, &arg) == -1)
      {
         perror("keyer.c keyRTS: RTS Clear failed");
      }
   }
   timer (ic_time);                      /* intra character time */
}


/*
 * setTone FUNCTION
 * sets sidetone freq
 */
void setTone (int new_tone)
{
   cw_set_frequency (new_tone);
   tone = 1193180 / new_tone;
}


/*
 * setSidetone FUNCTION
 * turns sidetones on or off
 */
void setSidetone (int tone_On)
{
   beepPC = tone_On & PC_TONE_ON ? 1 : 0;
   beepSC = tone_On & SC_TONE_ON ? 1 : 0;

   cw_set_console_file ("/dev/console");
   cw_set_console_sound (beepPC);
   cw_set_soundcard_sound (beepSC);
}


/*
 * setWPM FUNCTION
 */
void setWPM (int new_speed)
{
   struct tag1 reply;

   wpm = new_speed;
   cw_set_send_speed (wpm);

   dit_time = 1200000 / wpm;             /* dit time in usecs */
   dah_time = 3 * dit_time;
   ic_time = dit_time;                   /* intra letter time */

   setFarns (fwpm);                      /* set il and iw times with Farns */

   /* Tell the GUI we are done setting WPM */
   reply.type = WPM_DONE;
   reply.data.word[0] = ' ';
   reply.data.word[1] = (char) NULL;
   if (msgsnd (qid, (struct msgbuf *) &reply, sizeof (reply.data), 0) == -1)
   {
      perror ("keyer.c setWPM: msgsnd WPM_DONE failed");
      exit (1);
   }
}

/*
 * setFarns - set ic_time and il_time 
 */
void setFarns (int fwpm)
{
   long total;

   if (fwpm >= wpm || fwpm == 4)            /* use standard timing */
   {
      total = 0.0;
      il_time = 2 * dit_time;               /* 2 dits + ic_time */
      iw_time = 4 * dit_time;               /* 4 dits + il_time + ic_time */
   }
   else                                     /* use Farnsworth timing */
   {
      total = ((60 * wpm) - (37.2 *  fwpm)) / (wpm * fwpm) * 1000000;
      il_time = ((3 * total) / 19) - ic_time;
      iw_time = ((7 * total) / 19) - ic_time - il_time;
   }
}


/*
 * initialize FUNCTION
 */
void initialize (char *ser_dev)
{
   int i;

   /*
    * Open the serial port for keying and make sure the key is up
    */
   if (key_mode == TWCW)
   {
      if ((serfd = open (ser_dev, O_RDWR)) < 0)
      {
         perror ("keyer.c initialize: Can't open serial port");
         exit (1);
      }

      if (ioctl (serfd, TIOCMBIC, &arg) < 0)
      {
         perror("keyer.c initialize: RTS Clear failed");
         exit (1);
      }
   }

   /*
    * Initialize the letters array
    */
   for (i = 0; i < MAX_LETTERS; i++)
      letters[i] = (char) 0;

   /* Pro Signs and Special */   
   letters[1]         = "......";   /* F1 = Error */
   letters[2]         = ".-.-.";    /* F2 = <AR>  */
   letters[3]         = "...-.-";   /* F3 = <SK>  */
   letters[4]         = "-.--.";    /* F4 = <KN>  */
   letters[5]         = ".-...";    /* F5 = WAIT  */
   letters[6]         = "-...-.-";  /* F6 = BREAK */

   /* Numbers */
   letters[(int) '1'] = ".----";
   letters[(int) '2'] = "..---";
   letters[(int) '3'] = "...--";
   letters[(int) '4'] = "....-";
   letters[(int) '5'] = ".....";
   letters[(int) '6'] = "-....";
   letters[(int) '7'] = "--...";
   letters[(int) '8'] = "---..";
   letters[(int) '9'] = "----.";
   letters[(int) '0'] = "-----";
   letters[216]       = "-----";        /* 216 or 330h is a zero with a slash */

   /* Alphabet */
   letters[(int) 'A'] = ".-";
   letters[(int) 'B'] = "-...";
   letters[(int) 'C'] = "-.-.";
   letters[(int) 'D'] = "-..";
   letters[(int) 'E'] = ".";
   letters[(int) 'F'] = "..-.";
   letters[(int) 'G'] = "--.";
   letters[(int) 'H'] = "....";
   letters[(int) 'I'] = "..";
   letters[(int) 'J'] = ".---";
   letters[(int) 'K'] = "-.-";
   letters[(int) 'L'] = ".-..";
   letters[(int) 'M'] = "--";
   letters[(int) 'N'] = "-.";
   letters[(int) 'O'] = "---";
   letters[(int) 'P'] = ".--.";
   letters[(int) 'Q'] = "--.-";
   letters[(int) 'R'] = ".-.";
   letters[(int) 'S'] = "...";
   letters[(int) 'T'] = "-";
   letters[(int) 'U'] = "..-";
   letters[(int) 'V'] = "...-";
   letters[(int) 'W'] = ".--";
   letters[(int) 'X'] = "-..-";
   letters[(int) 'Y'] = "-.--";
   letters[(int) 'Z'] = "--..";

   /* Punctuation */
   letters[(int) '='] = "-...-";
   letters[(int) '/'] = "-..-.";
   letters[(int) '?'] = "..--..";
   letters[(int) ','] = "--..--";
   letters[(int) '.'] = ".-.-.-";
   letters[(int) ':'] = "---...";
   letters[(int) '\''] = ".----.";
   letters[(int) '"'] = ".-..-.";
   letters[(int) '_'] = "..--.-";
   letters[(int) '('] = "-.--.";
   letters[(int) ')'] = "-.--.-";
   letters[(int) '#'] = "-.---";
   letters[(int) '-'] = "-...-";
   letters[(int) '|'] = "...-..";
   letters[(int) '\\'] = "-.....";
   letters[(int) '*'] = "-----.";
   letters[(int) ';'] = "-.-.-.";
   letters[(int) '@'] = ".--.-.";
   letters[(int) '^'] = "....--.-.";
   letters[(int) '$'] = "...-..-";
   letters[(int) '!'] = "....-.";
   letters[(int) '>'] = "....---.";
   letters[(int) ']'] = "....-....";
   letters[(int) '['] = "....-..";
   letters[(int) '<'] = "....-.-..";
   letters[(int) '&'] = "....--.";
   letters[(int) '%'] = "....-.--.";
   letters[(int) '~'] = "....--";
   letters[(int) '+'] = ".-.-.";
   letters[(int) '{'] = "....-.--";
   letters[(int) '}'] = "....--..-";

   return;
}


/*
 * cleanup FUNCTION 
 * Make sure the sidetone is off and the key is up, then exit
 */
void cleanup (void)
{
   signal (SIGINT,  SIG_IGN);
   signal (SIGTERM, SIG_IGN);
   signal (SIGQUIT, SIG_IGN);
   signal (SIGSEGV, SIG_IGN);
   fprintf(stderr, "sendCW terminating\n");

   /* make sure sound is off and the key is up */
   if (confd != 0)
      ioctl (confd, KIOCSOUND, 0);

   if (serfd != 0)
      if (ioctl (serfd, TIOCMBIC, &arg) == -1)
         perror("keyer.c cleanup: RTS Clear failed"); 

   exit (73);
}

